/* PPProg - A Linux Device Driver for direct access to the
*           parallel port. (Kernel versions 2.0.x-2.4.x)
* Copyright (C) 2005 by Dirk Baechle (dl9obn@darc.de)
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public
* License along with this program; if not, write to the 
*
* Free Software Foundation, Inc.
* 675 Mass Ave
* Cambridge
* MA 02139
* USA
*
*/


/** \file ppprog.c
* Linux Device Driver for direct access to the parallel port.
* \author Dirk Baechle, dl9obn@darc.de
* \version 1.0
* \date 2005-05-12
*/

/* This file was created automatically from the file `ppprog.nw' by NOWEB.
* If you want to make changes, please edit the source `ppprog.nw'.
* A precompiled documentation can be found in `ppprog.pdf' and `ppprog.ps',
* respectively.
* Read it to understand why things are as they are. Thank you!
*/



/* Comment following define if ``parport'' */
/* driver should not be used. */
#define USE_PARPORT 1

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <asm/io.h>

/* Deal with CONFIG_MODVERSIONS */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

/* For character devices */
#include <linux/fs.h>
#include <linux/wrapper.h>

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>
#include <linux/delay.h>
#else
#include <asm/delay.h>
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Dirk Baechle");
MODULE_DESCRIPTION("Driver for direct access to the parallel port");
MODULE_SUPPORTED_DEVICE("ppprog");
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
MODULE_PARM (lp, "i");
MODULE_PARM (iobase, "i");
MODULE_PARM_DESC( lp, "Parallel port that should be used (0-3)" );
MODULE_PARM_DESC( iobase, "Parallel port i/o base address. Overrides `lp'. (default: 0x378)" );
#endif



#ifdef USE_PARPORT
#include <linux/parport.h>
#endif


#include "ppprog.h"


/** The value for success. */
#define SUCCESS                         0

/** The name of the device as it appears in /proc/devices. */
static char DEVICE_NAME[10] = "ppprog";

/** Is the device open? 1 equals yes, 0 equals no. */
static int device_is_open = 0;

/** Base address of the used parallel port (Default: 0x378) */
static int LptBase = 0x378;


#ifdef USE_PARPORT
/* Function prototypes */
void ppprog_attach(struct parport *);
void ppprog_detach(struct parport *);
/** Stores the pointers to the functions for attaching and detaching
* detected parallel ports. */
static struct parport_driver ppprog_driver = {
              DEVICE_NAME,
              ppprog_attach,
              ppprog_detach,
              NULL
};
#endif


#ifdef USE_PARPORT
/** Pointer to the struct of the allocated parallel port. */
struct parport *ppprog_port;
/** Pointer to the struct of the registered device, is needed
* for unregistering. */
struct pardevice *ppprog_device = 0;
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
/** Number of the parallel port that should be used (0-3),
* where ``0'' refers to \c LPT1. */
static int lp = -1;
/** Base IO port address of the parallel port. Overrides \a lp,
* if necessary. */
static int iobase = -1;
#endif


/** Attempts to open the device file.
* @param inode Pointer to the inode
* @param file Pointer to the device file
* @return 0 for success, else device is busy
*/
static int ppprog_open(struct inode *inode, struct file *file)
{
#if DEBUG
  printk(KERN_DEBUG "ppprog_open(%p, %p)\n", inode, file);
#endif

  
    /* Is the device open already? */
    if (device_is_open)
    {
  #if DEBUG
      printk(KERN_DEBUG "Device PPPROG is already opened!\n");
  #endif
      return(-EBUSY);
    }

  
  #ifdef USE_PARPORT
  
  if (parport_register_driver(&ppprog_driver) != 0)
  {
  #if DEBUG
      printk(KERN_DEBUG "PPPROG driver could not be registered with parport module!\n");
  #endif
      return(-EBUSY);
  }

  #else
    /* Is the region for the parallel port adapter still accessible? */
    if (check_region(LptBase, 3) != 0)
    {
  #if DEBUG
      printk(KERN_DEBUG "IO ports for parallel port adapter are not accessible!\n");
  #endif
      return(-EBUSY);
    }
  #endif

  
  #ifdef USE_PARPORT

    /* Get port with correct base number */
    ppprog_port = parport_find_base(LptBase);
    if (ppprog_port == NULL)
    {
  #if DEBUG
      printk(KERN_DEBUG "Parallel IO port %X could not be found!\n", LptBase);
  #endif
      return(-EBUSY);
    }

    ppprog_device = parport_register_device(ppprog_port,
                                          DEVICE_NAME,
                                          NULL,
                                          NULL,
                                          NULL,
                                          0,
                                          NULL);
    if (ppprog_device > 0)
    {
      if (parport_claim(ppprog_device) != 0)
      {
        parport_unregister_device(ppprog_device);
  #if DEBUG
        printk(KERN_DEBUG "Parallel IO port %X is not accessible via parport driver!\n", LptBase);
  #endif
        return(-EBUSY);
      }
    }
  #else
    /* Claim port regions */
    request_region(LptBase, 3, DEVICE_NAME);
  #endif


  device_is_open++;

  MOD_INC_USE_COUNT;

  return(SUCCESS);

}


/** Closes the device file.
* @param inode Pointer to the inode
* @param file Pointer to the device file
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int ppprog_release(struct inode *inode, struct file *file)
#else
static void ppprog_release(struct inode *inode, struct file *file)
#endif
{

#if DEBUG
  printk(KERN_DEBUG "ppprog_release(%p, %p)\n", inode, file);
#endif

  
  #ifdef USE_PARPORT
    parport_release(ppprog_device);
  #else
    /* Release port regions */
    release_region(LptBase, 3);
  #endif

  
  #ifdef USE_PARPORT
    parport_unregister_device(ppprog_device);
  #endif


  /* Release device counter */
  device_is_open--;

  MOD_DEC_USE_COUNT;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  return(0);
#endif

}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
/** Reads from the already opened device.
* @param file Pointer to the device file
* @param buffer Pointer to the buffer
* @param length Length of the buffer
* @param offset Offset to the file
* @return Number of bytes read
*/
static ssize_t ppprog_read(struct file *file, char *buffer, size_t length,
                         loff_t *offset)
#else
/** Reads from the already opened device.
* @param inode Pointer to inode
* @param file Pointer to the device file
* @param buffer Pointer to the buffer
* @param length Length of the buffer
* @return Number of bytes read
*/
static int ppprog_read(struct inode *inode, struct file *file, char *buffer,
                     int length)
#endif
{

#if DEBUG
  printk(KERN_DEBUG "ppprog_read(%p, %p, %p)\n", file, buffer, &length);
#endif

  put_user(inb(LptBase), buffer);

  return 1;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
/** Writes to the already opened device.
* @param file Pointer to the device file
* @param buffer Pointer to the buffer
* @param length Length of the buffer
* @param offset Offset to the file
* @return Number of bytes written
*/
static ssize_t ppprog_write(struct file *file, const char *buffer, size_t length,
                          loff_t *offset)
#else
/** Writes to the already opened device.
* @param inode Pointer to inode
* @param file Pointer to the device file
* @param buffer Pointer to the buffer
* @param length Length of the buffer
* @return Number of bytes written
*/
static int ppprog_write(struct inode *inode, struct file *file,
                      const char *buffer, int length)
#endif
{

#if DEBUG
  printk(KERN_DEBUG "ppprog_write (%p, %s, %d)", file, buffer, length);
#endif

  outb((char) (*buffer & 0xFF), LptBase);

  return 1;
}


/** Handles the ioctl calls of the device driver.
* @param inode Pointer to the inode
* @param file Pointer to the file
* @param ioctl_num Number of the ioctl
* @param ioctl_param Parameter, i.e. pointer to int
* @return 0
*/
int ppprog_ioctl(struct inode *inode, struct file *file,
               unsigned int ioctl_num, unsigned long ioctl_param)
{
  int *temp, cnt, commands;
  int data = 0;

  switch (ioctl_num)
  {
    
    case IOCTL_PPPROG_SEND:  temp = (int *) ioctl_param;
                             
                             #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
                             get_user(commands, temp++);
                             #else
                             commands = get_user(temp++);
                             #endif

                             for (cnt = 0; 
                                  ((cnt < MAX_ASSERTIONS) && (cnt < commands)); 
                                  cnt++)
                             {
                             #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
                               get_user(data, temp++);
                             #else
                               data = get_user(temp++);
                             #endif

                             #ifdef USE_PARPORT
                               parport_write_data(ppprog_device->port, (data & 0xFF));
                               parport_write_control(ppprog_device->port, ((data >> 16) & 0xFF));
                             #else
                               outb((char) (data & 0xFF), LptBase);
                               outb((char) ((data >> 16) & 0xFF), LptBase+2);
                             #endif

                             #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
                               get_user(data, temp++);
                             #else
                               data = get_user(temp++);
                             #endif

                               if (data > 0)
                                 udelay(data);

                             }


                             break;

    
    case IOCTL_PPPROG_CHECK: temp = (int *) ioctl_param;
                             
                             #ifdef USE_PARPORT
                             data = (parport_read_control(ppprog_device->port) << 16);
                             data |= (parport_read_status(ppprog_device->port) << 8);
                             data |= parport_read_data(ppprog_device->port);
                             #else
                             data = (inb(LptBase+2) & 0xFF) << 16;
                             data |= (inb(LptBase+1) & 0xFF) << 8;
                             data |= inb(LptBase) & 0xFF;
                             #endif
                             put_user(data, temp);

                             break;
  }

  return(SUCCESS);

}



/** Struct that holds the VFS functions for the device. */
static struct file_operations Fops =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
  owner: THIS_MODULE,
  read: ppprog_read,        /* read */
  write: ppprog_write,      /* write */
  ioctl: ppprog_ioctl,      /* ioctl */
  open: ppprog_open,        /* open */
  release: ppprog_release   /* release */
#else
  NULL,                     /* seek */
  ppprog_read,              /* read */
  ppprog_write,             /* write */
  NULL,                     /* readdir */
  NULL,                     /* select */
  ppprog_ioctl,             /* ioctl */
  NULL,                     /* mmap */
  ppprog_open,              /* open */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
  NULL,                     /* flush */
#endif
  ppprog_release            /* release */
#endif
};


/** Initializes the module by registering the device driver.
* @return 0 for success, < 0 for an error
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,13)
static int ppprog_init(void)
#else
int init_module(void)
#endif
{
  int ret;


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
  switch (lp)
  {
    case 0: LptBase = 0x378; break;
    case 1: LptBase = 0x278; break;
    case 2: LptBase = 0x378; break;
    case 3: LptBase = 0x378; break;
  }

  if (iobase > 0)
    LptBase = iobase;

#endif
  
  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    ret = register_chrdev(DEVICE_MAJOR, DEVICE_NAME, &Fops);
  #else
    ret = module_register_chrdev(0, DEVICE_NAME, &Fops);
  #endif

    /* Negative return values signify an error */
    if (ret < 0)
    {
      printk(KERN_ERR "PPPROG: <init_module> : Registering device failed with %d!", ret);
      return(ret);
    }

    printk(KERN_INFO "PPPROG: Device registered with major device number %d\n", DEVICE_MAJOR);


  return(SUCCESS);
}


/** Cleanup by unregistering the appropriate file from /proc
*/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,13)
static void ppprog_exit(void)
#else
void cleanup_module(void)
#endif
{
  int ret;

  
  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    ret = unregister_chrdev(DEVICE_MAJOR, DEVICE_NAME);
  #else
    ret = module_unregister_chrdev(DEVICE_MAJOR, DEVICE_NAME);
  #endif

    if (ret < 0)
    {
      printk(KERN_ERR "PPPROG: <cleanup_module> : Error %d while unregistering\n", ret);
    }


}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,13)
module_init(ppprog_init);
module_exit(ppprog_exit);
#endif



#ifdef USE_PARPORT

/** Attaches the found port to the device.
* @param port Pointer to struct for the found parallel port
*/
void ppprog_attach(struct parport *port)
{
  ;
}

/** Is called if a parallel port should be detached.
* @param port Pointer to struct for the parallel port
*/
void ppprog_detach(struct parport *port)
{
  ;
}


#endif

